import json
from google.cloud import secretmanager
from google.oauth2 import service_account
from google.cloud import bigquery

def get_gsm_secret(project_id, secret_name):
    '''Returnerer secret-verdien
    '''
    client = secretmanager.SecretManagerServiceClient()
    name = f"projects/{project_id}/secrets/{secret_name}/versions/latest"
    response = client.access_secret_version(request={"name": name})
    secret = json.loads(response.payload.data.decode('UTF-8'))
    return secret


def create_bigquery_client(project_id: str, secret_name_bigquery: str):
    """Lager en BigQuery client som kan hente data som dataframe.

    Parameters
    ----------
    project_id : str
        GSM project id for hemmeligheter
    secret_name_bigquery : str
        Hemmelighetens navn i GSM. Må være en BQ-sørvisbruker

    Examples
    --------
    >>> bq_client = create_bigquery_client(<id>, <secret_name>)
    >>> df = bq_client.query("select * from `{project_id_bq}.{datasett}.{kilde_tabell}`").to_dataframe()

    Returns
    -------
    google.cloud.bigquery.client.Client
        bigquery client som kan hente data som dataframe
    """    
    bq_secret = get_gsm_secret(project_id, secret_name_bigquery)
    creds = service_account.Credentials.from_service_account_info(bq_secret)
    bigquery_client = bigquery.Client(credentials=creds, project=creds.project_id)
    return bigquery_client



from google.cloud import secretmanager
from google.cloud.bigquery import QueryJobConfig
import json
from typing import Optional

from .bq_reader import BQReader
from .oracle_writer import OracleWriter

import logging


def get_secret_env(resource_name) -> dict:
    secrets = secretmanager.SecretManagerServiceClient()
    secret = secrets.access_secret_version(name=resource_name)
    secret_str = secret.payload.data.decode("UTF-8")
    return json.loads(secret_str)


class DataTransfer:
    def __init__(
        self,
        config,
        source_query,
        target_table=None,
        query_job_config: Optional[QueryJobConfig] = None,
    ):
        self.__config = config
        self.oracle_writer = OracleWriter(self.__config["oracle"], target_table=target_table)
        self.bq_reader = BQReader(
            self.__config["gcp"], source_query=source_query, query_job_config=query_job_config
        )

    def run(self, batch_limit=None, datatypes=None, convert_lists=False):
        for i, batch in enumerate(self.bq_reader):
            self.oracle_writer.write_batch(batch, datatypes=datatypes, convert_lists=convert_lists)
            if batch_limit:
                if i > batch_limit:
                    self.oracle_writer.cleanup()
                    break
            logging.info(f"total rows read: {self.bq_reader.total_rows_read}")
        self.oracle_writer.cleanup()



from .oracle_writer import OracleWriter

import pytest
import datetime


@pytest.fixture()
def batch():
    batch = []
    for i in range(2):
        batch.append(
            {
                "pk": i,
                "metadata": {"oppretted_tid": str(datetime.datetime.now())},
                "liste": ["ele1", "ele2"],
                "liste2": [{"key": "val{}".format(i)}],
            }
        )
        return batch


def test_convert_lists_and_dicts_in_batch_to_json(batch):
    OracleWriter.convert_lists_and_dicts_in_batch_to_json(batch)
    assert isinstance(batch[0]["metadata"], str)
    assert isinstance(batch[0]["liste"], str)
    assert isinstance(batch[0]["liste2"], str)



import json
from oracledb import connect


class OracleWriter:
    def __init__(self, config, target_table=None):
        self.__config = config
        self.con = connect(
            user=self.__config["DB_USER"],
            password=self.__config["DB_PASSWORD"],
            dsn=self.__config["DB_DSN"],
        )
        self.target_table = target_table or self.__config["target-table"]
        self.insert_string = None
        self.total_rows_inserted = 0
        self.execution_time = self.get_oracle_sysdate()

    def write_batch(self, batch, convert_lists=False, datatypes={}):
        if self.total_rows_inserted == 0:
            # self.prepare_table()
            pass
        if convert_lists:
            self.convert_lists_and_dicts_in_batch_to_json(batch)

        # Legger til lastet tid
        self.add_execution_time_to_batch(self.execution_time, batch)

        if not self.insert_string:
            self.create_insert_string(batch)
        with self.con.cursor() as cursor:
            try:
                if datatypes:
                    cursor.setinputsizes(**datatypes)
                cursor.executemany(self.insert_string, batch)
                self.total_rows_inserted += cursor.rowcount
            except Exception as e:
                self.cleanup(is_healthy=False)
                print(e)
                raise RuntimeError(e)
        self.con.commit()

    def cleanup(self, is_healthy=True):
        if is_healthy:
            self.con.commit()
        else:
            self.con.rollback()
        self.con.close()

    def prepare_table(self):
        with self.con.cursor() as cursor:
            cursor.execute(f"truncate table {self.target_table}")
        return True

    def get_oracle_sysdate(self):
        with self.con.cursor() as cursor:
            cursor.execute(f"select sysdate from dual")
            row = cursor.fetchone()
        return row[0]

    def create_insert_string(self, batch):
        column_names = batch[0].keys()
        self.insert_string = f"""
        insert into {self.target_table}
        ({', '.join(column_names)}) 
        values({', '.join([f':{col}' for col in column_names])})
        """
        return self.insert_string

    @staticmethod
    def convert_lists_and_dicts_in_batch_to_json(batch: list):
        """Looper gjennom alle dicts i batch og konverterer nesta lister og dicts til json"""
        for i, ele in enumerate(batch):
            for key in ele:
                if isinstance(ele[key], list):
                    ele[key] = json.dumps(ele[key])
                if isinstance(ele[key], dict):
                    ele[key] = json.dumps(ele[key])
            batch[i] = ele

    @staticmethod
    def add_execution_time_to_batch(time, batch: list):
        """legger til lastet_tid i batch"""
        for i, ele in enumerate(batch):
            ele["lastet_tid"] = time
            batch[i] = ele



import oracledb
import os
import pandas as pd

def _create_connection(secret):
    oracle_client = oracledb.connect(
        user=os.environ.get("DBT_ENV_SECRET_USER", secret.get("DB_USER", None)),
        password=os.environ.get("DBT_ENV_SECRET_PASS", secret.get("DB_PASSWORD", None)),
        dsn=secret.get("DB_DSN", None),
    )
    return oracle_client

def db_sql_run(sql_query, secret):
    oracle_client = _create_connection(secret)
    with oracle_client.cursor() as cursor:
        cursor.execute(sql_query)
        cursor.execute('commit')


def db_read_to_df(sql_query, secret, prefetch_rows = 1000):
    '''Function that returns the result of a sql query and the columns.
    '''
    oracle_client = _create_connection(secret)
    with oracle_client.cursor() as cursor:
        cursor.prefetchrows = prefetch_rows
        cursor.arraysize = prefetch_rows + 1
        cursor.execute(sql_query)
        rows = cursor.fetchall()
        cols = [col[0].lower() for col in cursor.description]
        return pd.DataFrame(rows, columns=cols)

def sql_df_to_db(sql_query, secret, val_dict):
    '''Insert data into database from a dataframe using sql query.
    '''
    oracle_client = _create_connection(secret)
    with oracle_client.cursor() as cursor:
        cursor.executemany(sql_query, val_dict, batcherrors=True, arraydmlrowcounts = False)
        print(f'cursor rowcount: {cursor.rowcount})')
        for error in cursor.getbatcherrors():
            print("Error", error.message, "at row offset", error.offset)
        cursor.execute('commit')




import os
import requests


def publish_docs(*, docs_url_project: str):
    """
    Funksjon for å publisere dbt-dokumentasjon til dbt-intern.nav.no

    Krever at 'dbt docs generate' er kjørt først, som lager de tre filene som blir
    lastet opp til dbt.intern.nav.no. Nettsiden er speilet på dbt.ansatt.nav.no

    Se dokumentasjon fra Nada på https://github.com/navikt/dbt-docs#publisering

    OBS! Ikke bruk underscore i docs_url_project, bruk heller bindestrek.

    Args:
        docs_url_project (str): siste ledd i url-en, feks 'dvh-aap'
    """    

    if docs_url_project is None:
        ...

        
    docs_url_complete = "https://dbt.intern.nav.no/docs/spenn/" + docs_url_project

    files = ["target/manifest.json", "target/catalog.json", "target/index.html"]
    multipart_form_data = {}
    for file_path in files:
        file_name = os.path.basename(file_path)
        with open(file_path, "rb") as file:
            file_contents = file.read()
            print(f"Gathering {file_path} ({len(file_contents)/1024:.0f} kB)")
            multipart_form_data[file_name] = (file_name, file_contents)

    res = requests.put(docs_url_complete, files=multipart_form_data)
    res.raise_for_status()

    print("HTTP PUT status: ", res.status_code, res.text)
    print()


if __name__ == "__main__":
    publish_docs()



# %%
from pathlib import Path
from yaml import safe_load
from dvh_tools.oracle import db_read_to_df
from dvh_tools.cloud_functions import get_gsm_secret
from dvh_tools.knada_vm_user import set_environ


def get_comments_from_oracle(
        *,
        project_id=None,
        secret_name=None,
        sources_yml_path="dbt/sources.yml"
        ):
    """
    Leser kildetabeller i sources.yml, kobler seg til Oracle, henter alle kommentarer
    og lager en 'comments_source.yml' som kan brukes i autogenerering til modeller.

    Oppdaterer/lager 'comments_source.yml' med tabell- og kolonnekommentarer.

    Antar at py-fila blir kjørt fra en mappe dbt-prosjektet, feks fra dbt/docs/
    Det er her ^ comments_source.yml (output) blir lagret.

    Args:
        project_id (str): GCP-prosjekt-ID. Defaults to None.
        secret_name (str): Hemmelighetsnavn i GSM. Defaults to None.
        sources_yml_path (str): Path til soruces.yml. Defaults to "dbt/sources.yml".
    
    Returns:
        None
    """
    # %%
    # henter hemmeligheter fra Google Secret Manager. trenger DSN
    print("setter hemmeligheter for Oracle tilkobling")
    if project_id is None or secret_name is None:
        print("Mangler prosjekt-ID og/eller hemmelighetsnavn")
        exit(1)
    secret_dict = get_gsm_secret(project_id, secret_name)
    set_environ()

    # %%
    # find the sources.yml file
    def find_project_root(current_path):
        """Recursively find the project's root directory by looking for a specific marker (e.g., '.git' folder)."""
        if (current_path / '.git').exists():
            return current_path
        else:
            return find_project_root(current_path.parent)

    def find_all_sources_from_yml(sources_yml_path=sources_yml_path):
        """Finner alle kilder fra sources.yml."""
        print("Finner sources.yml fra:", sources_yml_path)
        project_root = find_project_root(Path(__file__).resolve())
        source_file = project_root / sources_yml_path  # Adjust this line if sources_yml_path should not be relative to project_root
        try:
            with open(source_file, "r") as file:
                content = file.read()
        except FileNotFoundError:
            print(f"Finner ikke yaml-filen hvor sources er spesifisert")
            print(f"Prøvde å lese fra: {source_file}")
            print(f"Endre argumentet 'sources_yml_path' til riktig path, som nå er: {sources_yml_path}")
            exit(1)
        yml_raw = safe_load(content)
        schema_list = yml_raw["sources"]
        schema_table_dict = {}  # schema som key, liste av tabellnavn som value
        for schema in schema_list:
            if schema["name"] != schema["schema"]:
                print("Obs! verdiene for name og schema er ulike! Se:", schema)
            schema_name = schema["name"]
            tables_name_list = []
            for table in schema["tables"]:
                tables_name_list.append(table["name"])
            schema_table_dict[schema_name] = tables_name_list
        return schema_table_dict
    schema_table_dict = find_all_sources_from_yml()

    # %%
    # sql-er mot Oracle for tabell- og kolonnekommentarer
    def sql_table_comment(schema_name: str, table_name: str) -> str:
        """Henter tabellkommentar fra Oracle-databasen.
        Args:
            schema_name (str): skjemanavn
            table_name (str): tabellnavn
        Returns:
            str: tabellkommentaren"""
        sql = f"""select comments from all_tab_comments
            where owner = upper('{schema_name}') and table_name = upper('{table_name}')"""
        sql_result = db_read_to_df(sql, secret_dict)
        if sql_result.empty or sql_result.iloc[0, 0] is None:
            return " "
        else:
            # fjerner fnutter, fordi det skaper problemer senere
            return sql_result.iloc[0, 0].replace("'", "").replace('"', "")

    def sql_columns_comments(schema_name: str, table_name: str) -> dict:
        """Henter alle kolonnekommentarer til en tabell i databasen.
        Args:
            schema_name (str): skjemanavn
            table_name (str): tabellnavn
        Returns:
            pd.dataframe: df med 'column_name' og 'comments'"""
        sql = f"""select column_name, comments from dba_col_comments
            where owner = upper('{schema_name}') and table_name = upper('{table_name}')"""
        df_col_comments = db_read_to_df(sql, secret_dict)
        df_col_comments["column_name"] = df_col_comments["column_name"].str.lower()
        df_col_comments["comments"] = df_col_comments["comments"].str.replace("'", "").str.replace('"', "")
        df_col_comments["comments"] = df_col_comments["comments"].fillna(" ")
        return df_col_comments

    # %%
    # get table descriptions
    print("Henter tabellbeskrivelser fra Oracle")
    src_table_descriptions = {}  # kommentar til source-folder
    stg_table_descriptions = {}  # kommentar til staging-modeller
    for schema, table_list in schema_table_dict.items():
        for table in table_list:
            source_description = sql_table_comment(schema, table).replace("\n", " | ")
            if source_description is None:
                source_description = "(ingen modellbeskrivelse i Oracle)"
            stg_table_descriptions[f"stg_{table}"] = f"Staging av {schema}.{table}, med original beskrivelse: {source_description}"
            src_table_descriptions[table] = source_description

    # %%
    # makes the file dbt/models/sources_with_comments.yml
    # and fills in the dict with unique column comments
    print("Henter kolonnekommentarer fra Oracle")
    print("Lager 'sources_with_comments.yml'")
    column_comments_dict = {}
    yml = "# IKKE ENDRE DENNE FILA!\n"
    yml += "# Den er autogenerert av dvh_tools.dbt.tools.get_comments_from_oracle\n"
    yml += "# Fjern/legg til kilder i dbt/sources.yml\n\n"
    yml += """version: 2\n\nsources:\n"""
    for schema, table_list in schema_table_dict.items():
        yml += f"  - name: {schema}\n"
        yml += f"    schema: {schema}\n"
        yml += f"    tables:\n"
        for table in table_list:
            yml += f"      - name: {table}\n"
            yml += f"        description: '{src_table_descriptions[table]}'\n"
            yml += f"        columns:\n"
            df_table_columns_comments = sql_columns_comments(schema, table)
            for _, row in df_table_columns_comments.iterrows():
                yml += f"          - name: {row['column_name']}\n"
                yml += f"            description: '{row['comments'].replace("\n"," | ")}'\n"
                # get unique column comments
                column = row["column_name"]
                comment = row["comments"]
                if column not in column_comments_dict:
                    column_comments_dict[column] = comment
    column_comments_dict = dict(sorted(column_comments_dict.items()))

    # lage source_comments.yml, med kolonnekommentarer og staging-kommentarer
    print("Lager 'comments_source.yml'")
    alle_kommentarer = "{\n    source_column_comments: {\n"
    for column, comment in column_comments_dict.items():
        alle_kommentarer += f"""        {column}: "{comment.replace('\n', " | ")}",\n"""
    alle_kommentarer += "    },\n\n    source_table_descriptions: {\n"
    for table, description in stg_table_descriptions.items():
        alle_kommentarer += f"""        {table}: "{description}",\n"""
    alle_kommentarer += "    }\n}\n"

    project_root = find_project_root(Path(__file__).resolve())
    with open(project_root / "dbt/models/sources_with_comments.yml", "w") as file:
        file.write(yml)
    with open(project_root / "dbt/docs/comments_source.yml", "w") as file:
        file.write(alle_kommentarer)
    print("Ferdig!")


from .dbt_docs_publish import publish_docs
from .get_comments_from_oracle import get_comments_from_oracle
from .generate_comments_from_sql import generate_comments_from_sql

__all__ = ["publish_docs", "get_comments_from_oracle", "generate_comments_from_sql"]


# setter bruker og passord for lokal kjøring i knada VM
# setter miljøvariablene DBT_ENV_SECRET_USER og DBT_ENV_SECRET_PASS, som altså blir lagret i minnet

import os
from getpass import getpass

# Brukernavn kan settes manuelt eller fra denne listen, bare å legge til flere
user_dict = {
    '1': 'M167094',
    '2': 'W158886',
    '3': 'M167094[DVH_AAP]',
    '4': 'W158886[DVH_AAP]',
    '5': 'M167094[DVH_TILLEGGSSTONADER]',
}

def get_environ_input():
    print("Hurtigvalg for brukere:")
    for key, value in user_dict.items():
        print(f"    {key}: {value}")
    user = input("Velg bruker med nummer eller skriv inn manuelt: ")
    if user in user_dict:
        user = user_dict[user]
    password = getpass("Passord: ")
    return user, password


def set_environ():
    """Bruker input og getpass for å sette miljøvariablene USER og PASSWORD.
    
    OBS! Dette funker i en jupyter notebook, men blir ikke lagret i minnet når python scriptet avsluttes.
    Bruk da heller bash-scriptet environment_local_user.sh for å sette miljøvariablene."""    
    user, password = get_environ_input()
    os.environ['DBT_ENV_SECRET_USER'] = user
    os.environ['DBT_ENV_SECRET_PASS'] = password
    print("Miljøvariablene DBT_ENV_SECRET_USER og DBT_ENV_SECRET_PASS satt for: ", os.environ.get("DBT_ENV_SECRET_USER"))


if __name__ == "__main__":
    set_environ()



import os
from yaml import safe_load

def make_yml_string(yml: dict) -> str:
    """Using a yml dict to make a string to write to a yml file.

    Args:
        yml (dict): A yml dict, must have 'version' and 'models'

    Returns:
        str: A string to write to a yml file
    """
    # antar at yml er en dict med 'version' og 'models'
    # lager output string til 1 yml-fil
    fallback_v = "2"
    try:
        yml_version = yml["version"]
    except KeyError:
        print(f"Ingen 'version' i yml-filen. Bruker fallback-versjon: {fallback_v}")
        yml_version = fallback_v
    yml_string = f"version: {yml_version}\n\nmodels:\n"

    # loop over tabeller
    for tab in yml["models"]:
        tab_keys = tab.keys()
        yml_string += f"  - name: {tab['name']}\n"  # må ha name
        indent_4 = "    "  # for tabell
        indent_6 = "      "  # for konfig til tabell og kolonner
        indent_8 = "        "  # for konfig kolonner
        indent_10 = "          "  # for konfig kolonner med lister/dict
        for key in tab_keys:
            if key == "name" or key == "columns":
                continue
            elif key == "description":
                yml_string += f"{indent_4}{key}: >\n{indent_6}{tab[key].strip()}\n"
            elif type(tab[key]) == str:
                yml_string += f"{indent_4}{key}: {tab[key].strip()}\n"
            elif type(tab[key]) == list:
                yml_string += f"{indent_4}{key}:\n"
                for list_item in tab[key]:
                    yml_string += f"{indent_6}- {list_item}\n"
            elif type(tab[key]) == dict:
                yml_string += f"{indent_4}{key}:\n"
                for ik, iv in tab[key].items():
                    yml_string += f"{indent_6}{ik}: {iv}\n"
            else:
                print(f"Ukjent type for {key} i {tab['name']}. Type: {type(tab[key])}")

        # loop over kolonner
        yml_string += indent_4 + "columns:\n"
        for col in tab["columns"]:
            yml_string += f"{indent_6}- name: {col['name']}\n"
            for ckey in col.keys():
                if ckey == "name":
                    continue
                elif ckey == "description":
                    yml_string += f"{indent_8}description: '{col['description']}'\n"
                elif type(col[ckey]) == str:
                    yml_string += f"{indent_8}{ckey}: {col[ckey].strip()}\n"
                elif type(col[ckey]) == list:
                    yml_string += f"{indent_8}{ckey}:\n"
                    for col_list_item in col[ckey]:
                        yml_string += f"{indent_10}- {col_list_item}\n"
                elif type(col[ckey]) == dict:
                    yml_string += f"{indent_8}{ckey}:\n"
                    for ik, iv in col[ckey].items():
                        yml_string += f"{indent_10}{ik}: {iv}\n"
                else:
                    print(f"Ukjent type for {col} i {tab['name']}. Type: {type(col[ckey])}")
        yml_string += "\n"
    return yml_string


def find_sql_columns(file) -> list:
    # returns a list of the columns in the sql-file
    with open(file, "r") as file:
        content = file.readlines()

    model_columns = []
    # two alternatives:
    ### 1. the with clause, finding "final as(\n"  # todo: add support leading comma
    ### 2. flat select statements, finding "select\n"
    try:
        # 1. with clause
        if content[-1].strip() == "select * from final":
            # find the lines between "    select" and "    from ..."
            select_line = content.index("final as (\n")
            read_from_index = select_line + 2
        else:  # flat select
            select_line = content.index("select\n")
            read_from_index = select_line + 1

        # todo: funker ikke å splitte på "." hvis det er en kommentar på linja
        #     column = column.lower()
        #     if "." in column:  # search for ".", if the column is aliased
        #         column = column.split(".")[1]

        for column in content[read_from_index:]:
            if column.strip().startswith("from"):
                break  # stop when reaching "from" in the sql-file
            elif column.strip().startswith("--"):
                continue  # skip commented lines
            elif column.strip().startswith("*"):
                print(f"\nError reading {file.name}")
                print("Do not end with 'select *' statements")
                print("Finish with explicit 'final as(' statement or a flat select")
                exit()
            elif column.count("--") > 0:
                # if the column has a comment, split on the first "--"
                column_name = column.split("--")[0].strip().replace(",", "")
                model_columns.append(column_name)
            else:
                try:  # when aliasing
                    column.split(" as ")[1]
                    column_name = column.split(" as ")[1].strip().replace(",", "")
                    model_columns.append(column_name)
                except IndexError:  # all normal columns
                    column_name = column.strip().replace(",", "")
                    model_columns.append(column_name)
    except ValueError:
        print(f"\nError reading {file.name}")
        print("Make sure to follow the standard structure of the sql-files,")
        print("i.e. use the with clause and 'final as(', or flat select statements")
        exit()
    return model_columns


def remove_column(yml: dict, model_name: str, column_name: str):
    # remove column from yml
    for i, mod in enumerate(yml["models"]):
        if mod["name"] == model_name:
            for j, col in enumerate(mod["columns"]):
                if col["name"] == column_name:
                    yml["models"][i]["columns"].pop(j)
                    break
            break


def add_column_empty_description(yml: dict, model_name: str, column_name: str):
    # add column to yml
    for i, mod in enumerate(yml["models"]):
        if mod["name"] == model_name:
            yml["models"][i]["columns"].append({"name": column_name, "description": ""})
            break


def empty_model_dict(model_name: str):
    return {"name": model_name, "description": "", "columns": []}


def update_yml_dict(*, yml_dict: dict, sql_dict: dict, yml_file: str) -> None:
    """
    Updates the yml dict by adding or removing models and columns.

    Args:
        yml_dict (dict): dict from the yml file
        sql_dict (dict): dict from the sql files, with models as keys and columns as values
        yml_file (str): file name of the yml file
    """
    yml_mod_names = [model["name"] for model in yml_dict["models"]]

    # new sql model
    for sql_model in sql_dict:
        if sql_model in yml_mod_names:
            continue
        else:
            print(f"Appending {sql_model} to {yml_file} with empty desc")
            yml_dict["models"].append(empty_model_dict(sql_model))

    # model in yml but not in sql
    for i, yml_model_n in enumerate(yml_mod_names):
        if yml_model_n not in sql_dict:
            print(f"Popping model {yml_model_n} from {yml_file}")
            yml_dict["models"].pop(i)

    # updating the columns
    for model in yml_dict["models"]:
        model_name = model["name"]
        model_cols = model["columns"]
        model_col_names = [col["name"] for col in model_cols]
        for col in model_cols:
            if col["name"] in sql_dict[model_name]:
                continue
            # column not in sql
            else:
                print(f"Popping {col['name']} from {model_name} in {yml_file}")
                remove_column(yml_dict, model_name, col["name"])
        # column in sql but not in yml
        for sql_col in sql_dict[model_name]:
            if sql_col not in model_col_names:
                print(f"Appending {sql_col} to {model_name}")
                add_column_empty_description(yml_dict, model_name, sql_col)


def update_yml_in_dir(files_and_dirs: list, model_dir: str, models_path: str = None) -> None:
    sql_files = [f for f in files_and_dirs if f.endswith(".sql")]
    yml_file = [f for f in files_and_dirs if f.endswith(".yml")]
    if "sources.yml" in yml_file:
        yml_file.remove("sources.yml")

    sql_dict = {}  # models as keys, columns as values
    if len(sql_files) > 0:
        for file in sql_files:
            file_name = file[: -len(".sql")]
            with open(models_path + model_dir + "/" + file, "r") as f:
                model_columns = find_sql_columns(models_path + model_dir + "/" + file)
                sql_dict[file_name] = model_columns

    if len(yml_file) > 0:
        with open(models_path + model_dir + "/" + yml_file[0], "r") as f:
            yml_dict = safe_load(f)
        try:
            yml_models_dict = yml_dict["models"]
        except KeyError:
            print(f"No 'models' in {yml_file[0]}")
            yml_models_dict = None

        if yml_models_dict:
            update_yml_dict(yml_dict=yml_dict, sql_dict=sql_dict, yml_file=yml_file[0])
            yml_string = make_yml_string(yml_dict)
            with open(models_path + model_dir + "/" + yml_file[0], "w") as f:
                f.write(yml_string)

    # hvis det ikke er noen yaml, men det er sql-filer
    if len(yml_file) == 0 and len(sql_files) > 0:
        print(f"No yml file in {model_dir}, but found {sql_files}. Making the yml file.")
        # first make a dummy yml dict
        yml_dict = {"version": "2", "models": [{"name": "dummy", "columns": [{"name": "aarmnd"}]}]}
        update_yml_dict(yml_dict=yml_dict, sql_dict=sql_dict, yml_file="dummy.yml")
        yml_string = make_yml_string(yml_dict)
        new_file_name = "/_" + model_dir.split("/")[-1] + "_models.yml"
        with open(models_path + model_dir + new_file_name, "w") as f:
            f.write(yml_string)


def update_yml_from_sql(*, models_path: str):
    """
    Oppdaterer .yml-filene i dbt-prosjektet med kolonner fra .sql-filene.
    De får tomme kommentarer med denne funksjonen, men det er midlertidig.
    Kommentarene blir fylt ut i neste steg, men må være i .ym-filene for å kunne fylles ut.
    """

    # looping over the dbt models dir and subdirs
    for model_dir in ["staging", "marts", "intermediate"]:
        # first look for .sql files in the directory
        files_and_dirs = os.listdir(models_path + model_dir)
        update_yml_in_dir(files_and_dirs, model_dir, models_path)
        # then look for subdirs
        subdirs = [d for d in files_and_dirs if os.path.isdir(models_path + model_dir + "/" + d)]
        for subdir in subdirs:
            files_and_dirs = os.listdir(models_path + model_dir + "/" + subdir)
            update_yml_in_dir(files_and_dirs, model_dir + "/" + subdir, models_path)



"""
Gather all various operations that can be re-used.
"""


def convert_listtuples_tuple(self, result):
    '''
    Read: Convert "list of tuples" returned by oracle query to "tuple"
    Example: convert_listtuples_tuple(result from oracle cursor)
    '''
    return next(zip(*result))


import glob
from pathlib import Path
from yaml import safe_load
from dvh_tools.dbt_tools.generate_comments_utils import update_yml_from_sql, make_yml_string

def generate_comments_from_sql(*, models_path="dbt/models", docs_path="dbt/docs"):
    # update_yml_from_sql oppdaterer yml-filene i henhold til sql-filene
    # i.e. fjerner/legger til kolonner/modeller basert på sql-filstrukturen
    def find_project_root(current_path):
        """Recursively find the project's root directory by looking for a specific marker (e.g., '.git' folder)."""
        if (current_path / '.git').exists():
            return current_path
        else:
            return find_project_root(current_path.parent)
    project_root = find_project_root(Path(__file__).resolve())
    models_path = str(project_root / models_path) + "/"
    yaml_files = glob.glob(models_path + "**/*.yml", recursive=True)

    update_yml_from_sql(models_path=models_path)

    overskriv_yml_med_custom = True  # overskriving av det i yml-filene med custom_comments
    endre_bare_tomme_kommentarer = False  # endrer bare tomme kommentarer, eller alle

    column_descriptions = {}
    table_descriptions = {}


    try:  # lese custom comments
        with open(str(project_root / docs_path / "comments_custom.yml")) as f:
            custom_comments = safe_load(f)
            custom_column_comments = custom_comments["custom_column_comments"]
            custom_table_descriptions = custom_comments["custom_table_descriptions"]
    except Exception as e:
        print(e)
        print("Ha en fil med kommentarer i 'comments_custom.yml'")

    try:  # lese source_column_comments
        with open(str(project_root / docs_path / "comments_source.yml")) as f:
            source_comments = safe_load(f)
            source_column_comments = source_comments["source_column_comments"]
            source_table_descriptions = source_comments["source_table_descriptions"]
            table_descriptions.update(source_table_descriptions)
    except Exception as e:
        print(e)
        print("Fant ikke 'comments_source.yml, som skal ha kommentarer fra source'")

    # først samle inn alle kolonnenavn og beskrivelser
    kolonner_navn = []
    kolonner_kommentar = []
    for file in yaml_files:
        # hvis fila er "sources.yml", hopp over
        if "/sources.yml" in file or "/sources_with_comments.yml" in file:
            continue
        with open(file, "r") as f:
            yml = safe_load(f)
            try:
                tabeller = yml["models"]
            except KeyError:
                print(f"KeyError on 'models' in {file}")
                continue
            for t in tabeller:
                t_name = t["name"]
                t_columns = t["columns"]
                if "description" in t:
                    table_descriptions[t_name] = t["description"]
                for c in t_columns:
                    c_name = c["name"]
                    try:
                        c_description = c["description"]
                    except KeyError:
                        print(f"{c_name} har ikke felt for beskrivelse i {t_name}")
                        continue
                    if c_description is None or c_description == "":
                        # print(f"{c_name} har ingen/tom beskrivelse i {t_name}")
                        continue
                    if c_name in kolonner_navn:
                        continue  # henter kun unike kolonnenavn og første beskrivelse
                    else:
                        kolonner_navn.append(c_name)
                        kolonner_kommentar.append(c_description)
    yml_column_comments = dict(zip(kolonner_navn, kolonner_kommentar))

    # custom > yml > source
    # overskriver source_column_comments med yml_column_comments
    for col, desc in source_column_comments.items():
        column_descriptions[col] = desc
    # overskriv databasebeskrivelser med yml
    column_descriptions.update(yml_column_comments)
    # eventuelt oppdater med custom_column_comments
    if overskriv_yml_med_custom:
        column_descriptions.update(custom_column_comments)
    # legge til nye column comments
    for col, desc in custom_column_comments.items():
        column_descriptions[col] = desc
    table_descriptions.update(custom_table_descriptions)

    manglende_kommentarer = []
    # Så parse filene og smelle inn nye kommentarer
    for f in yaml_files:
        # hvis fila er "sources.yml", hopp over
        if "/sources.yml" in f or "/sources_with_comments.yml" in f:
            continue
        with open(f, "r") as file:
            yml = dict(safe_load(file))
            yml_models = False
            try:
                yml["models"].sort(key=lambda x: x["name"])
                tabeller = yml["models"]
                yml_models = True
            except KeyError:
                print(f"Ingen 'models' i .yml {f}")
                continue
            if yml_models:
                # loop over dbt modeller i yml-fila
                for i in range(len(tabeller)):
                    t_name = tabeller[i]["name"]
                    t_columns = tabeller[i]["columns"]
                    if "description" in tabeller[i]:
                        t_desc = tabeller[i]["description"]
                        if t_desc.strip() != table_descriptions[t_name].strip():
                            print(f"Endrer beskrivelse for modell {t_name}")
                            yml["models"][i]["description"] = table_descriptions[t_name]
                    # loop over kolonnene i en modell
                    for c in range(len(t_columns)):
                        c_name = t_columns[c]["name"]
                        overskriv_beskrivelse = False
                        if not endre_bare_tomme_kommentarer:
                            overskriv_beskrivelse = True
                        try:
                            c_desc = t_columns[c]["description"]
                        except KeyError:  # ingen beskrivelse av kolonnen
                            overskriv_beskrivelse = True
                            c_desc = None
                        if c_name not in column_descriptions:
                            # print(f"Fant ingen beskrivelse å bruke for {c_name}")
                            overskriv_beskrivelse = False  # får ikke overskrevet
                            if c_name not in manglende_kommentarer:
                                manglende_kommentarer.append(c_name)
                        if overskriv_beskrivelse and c_desc != column_descriptions[c_name]:
                            print(f"Endrer beskrivelse for {c_name} i {t_name}")
                            oppdatert_desc = column_descriptions[c_name]
                            yml["models"][i]["columns"][c]["description"] = oppdatert_desc

        # skriver hver enkelt .yml-fil
        with open(f, "w") as file:
            file.write(make_yml_string(yml))

    if len(manglende_kommentarer) > 0:
        print("mangler følgende kolonner i comments_custom.yml:")
        for c_name in manglende_kommentarer:
            print("   ", c_name)



from typing import Optional
from google.cloud import bigquery
from google.oauth2 import service_account


class BQReader:
    def __init__(
        self, config, source_query, query_job_config: Optional[bigquery.QueryJobConfig] = None
    ):
        self.__config = config
        self.client = bigquery.Client(
            credentials=service_account.Credentials.from_service_account_info(self.__config)
        )
        self._query_job = self.client.query(source_query, job_config=query_job_config)

        self._generator = self.__batch_generator()
        self.total_rows_read = 0

    def __iter__(self):
        return self

    def __next__(self):
        rows = [dict(row) for row in next(self._generator)]
        if len(rows) > 0:
            self.total_rows_read += len(rows)
            return rows
        else:
            raise StopIteration

    def __batch_generator(self):
        for page in self._query_job.result(page_size=500000).pages:
            yield page
